/* File "malrul.c":
 * Takes a rule file and compiles it to a intermediate code file. */

/* This file is part of Malaga, a system for Left Associative Grammars.
 * Copyright (C) 1995-1998 Bjoern Beutel
 *
 * Bjoern Beutel
 * Universitaet Erlangen-Nuernberg
 * Abteilung fuer Computerlinguistik
 * Bismarckstrasse 12
 * D-91054 Erlangen
 * e-mail: malaga@linguistik.uni-erlangen.de 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "basic.h"
#include "pools.h"
#include "values.h"
#include "symbols.h"
#include "scanner.h"
#include "files.h"
#include "instr_type.h"
#include "rule_code.h"
#include "rule_compiler.h"

#ifdef HANGUL
#include "hangul.h"
#endif

#undef GLOBAL
#define GLOBAL

/*---------------------------------------------------------------------------*/

LOCAL string_t program_name; /* the name the program has been called by */

/*---------------------------------------------------------------------------*/

GLOBAL void error (string_t format, ...)
/* Print an error message and exit. */
{
  va_list arg;
  
  fprintf (stderr, "%s: ", program_name);

  if (current_line_number () != -1) 
    fprintf (stderr, "file \"%s\", line %ld, column %ld: ", 
	     name_in_path (current_file_name ()), current_line_number (), 
	    current_column ());

  va_start (arg, format);
  vfprintf (stderr, format, arg);
  va_end (arg);

  fprintf (stderr, "\n");

  if (getenv ("MALAGA_MODE") != NULL && current_line_number () != -1)
    printf ("SHOW \"%s\":%ld:%ld\n", 
	    current_file_name (), current_line_number (), current_column ());

  exit (1);
}

/*---------------------------------------------------------------------------*/

LOCAL void program_message (void)
/* Print some information about the program. */
{
  printf ("%s (%s) - Copyright (C) 1995-1998 Bjoern Beutel\n",
	  program_name, MALAGA_VERSION);
  printf ("This program comes with ABSOLUTELY NO WARRANTY.\n");
  printf ("This is free software which you may redistribute "
	  "under certain conditions.\n");
  printf ("For details, refer to the GNU General Public License.\n");
}

/*---------------------------------------------------------------------------*/

GLOBAL int main (int argc, string_t argv[])
/* The main function of "malrul". */
{
  string_t rule_file, object_file, symbol_file;
  short_t file_type, i;

  /* Prevent warning about uninitialised variable. */
  file_type = 0;

  program_name = name_in_path (argv[0]);

#ifdef HANGUL
  init_hangul ();
#endif

  if (argc == 2 && (strcmp_no_case (argv[1], "-version") == 0
		    || strcmp_no_case (argv[1], "-v") == 0))
  {
    program_message ();
    exit (0);
  }

  rule_file = object_file = symbol_file = NULL;
  for (i = 1; i < argc; i++) 
  {
    string_t argument = ((*argv[i] == '-') 
			 ? argv[i] : absolute_path (argv[i], NULL));
    
    if (has_extension (argument, "all")) 
    {
      set_file_name (&object_file, argument, "all_c");
      rule_file = new_string (argument);
      file_type = ALLO_RULE_FILE;
    } 
    else if (has_extension (argument, "mor")) 
    {
      set_file_name (&object_file, argument, "mor_c");
      rule_file = new_string (argument);
      file_type = MORPHO_RULE_FILE;
    } 
    else if (has_extension (argument, "mfil")) 
    {
      set_file_name (&object_file, argument, "mfil_c");
      rule_file = new_string (argument);
      file_type = FILTER_RULE_FILE;
    } 
    else if (has_extension (argument, "ifil")) 
    {
      set_file_name (&object_file, argument, "ifil_c");
      rule_file = new_string (argument);
      file_type = FILTER_RULE_FILE;
    } 
    else if (has_extension (argument, "syn")) 
    {
      set_file_name (&object_file, argument, "syn_c");
      rule_file = new_string (argument);
      file_type = SYNTAX_RULE_FILE;
    } 
    else if (has_extension (argument, "sfil")) 
    {
      set_file_name (&object_file, argument, "sfil_c");
      rule_file = new_string (argument);
      file_type = FILTER_RULE_FILE;
    } 
    else if (has_extension (argument, "nav")) 
    {
      set_file_name (&object_file, argument, "nav_c");
      rule_file = new_string (argument);
      file_type = SYNTAX_RULE_FILE;
    } 
    
    else if (has_extension (argument, "sym_c"))
      set_file_name (&symbol_file, argument, NULL);
    else if (has_extension (argument, "esym_c"))
      set_file_name (&symbol_file, argument, NULL);
    else 
      error ("illegal argument \"%s\"", argument);
  }
  
  if (object_file == NULL)
    error ("no file to compile");
  
  if (symbol_file == NULL)
    error ("missing symbol file name");

  read_symbol_file (symbol_file);

  compile_rule_file (rule_file, object_file, file_type);
    
  free_symbol_table ();
  free_value_heap ();
  free (symbol_file);
  free (object_file);
  free (rule_file);

#ifdef HANGUL
  terminate_hangul ();
#endif

  return 0;
}

/*---------------------------------------------------------------------------*/
